package io.gitlab.jfronny.libjf.data.manipulation.api;

import io.gitlab.jfronny.commons.LazySupplier;
import io.gitlab.jfronny.commons.throwable.*;
import io.gitlab.jfronny.libjf.LibJf;
import io.gitlab.jfronny.libjf.data.manipulation.impl.ResourcePackHook;
import net.fabricmc.fabric.api.event.Event;
import net.fabricmc.fabric.api.event.EventFactory;
import net.minecraft.resource.ResourcePack;
import net.minecraft.resource.ResourceType;
import net.minecraft.util.Identifier;

import java.io.IOException;
import java.io.InputStream;
import java.util.Collection;
import java.util.function.Predicate;
import java.util.function.Supplier;

@SuppressWarnings("unused")
public class UserResourceEvents {
    public static <TVal, TEx extends Throwable> TVal disable(ThrowingSupplier<TVal, TEx> then) throws TEx {
        try {
            ResourcePackHook.setDisabled(true);
            return then.get();
        }
        finally {
            ResourcePackHook.setDisabled(false);
        }
    }

    public static <TEx extends Throwable> void disable(ThrowingRunnable<TEx> then) throws TEx {
        try {
            ResourcePackHook.setDisabled(true);
            then.run();
        } finally {
            ResourcePackHook.setDisabled(false);
        }
    }

    public static final Event<Contains> CONTAINS = EventFactory.createArrayBacked(Contains.class,
            (listeners) -> (type, id, previous, pack) -> {
                LazySupplier<Boolean> lazy = new LazySupplier<>(previous);
                for (Contains listener : listeners) {
                    lazy = lazy.andThen(supplier -> listener.contains(type, id, supplier, pack));
                }
                return lazy.get();
            });

    public static final Event<FindResource> FIND_RESOURCE = EventFactory.createArrayBacked(FindResource.class,
            (listeners) -> ((type, namespace, prefix, allowedPathPredicate, previous, pack) -> {
                LazySupplier<Collection<Identifier>> lazy = new LazySupplier<>(previous);
                for (FindResource listener : listeners) {
                    lazy = lazy.andThen(supplier -> listener.findResources(type, namespace, prefix, allowedPathPredicate, supplier, pack));
                }
                return lazy.get();
            }));

    public static final Event<Open> OPEN = EventFactory.createArrayBacked(Open.class,
            (listeners) -> ((type, id, previous, pack) -> {
                LazySupplier<InputStream> lazy = new LazySupplier<>(previous);
                for (Open listener : listeners) {
                    lazy = lazy.andThen(supplier -> {
                        try {
                            return listener.open(type, id, supplier, pack);
                        } catch (IOException e) {
                            LibJf.LOGGER.error("Could not call ResourcePack.OPEN listener", e);
                            return null;
                        }
                    });
                }
                return lazy.get();
            }));

    public static final Event<OpenRoot> OPEN_ROOT = EventFactory.createArrayBacked(OpenRoot.class,
            (listeners) -> ((fileName, previous, pack) -> {
                LazySupplier<InputStream> lazy = new LazySupplier<>(previous);
                for (OpenRoot listener : listeners) {
                    lazy = lazy.andThen(supplier -> {
                        try {
                            return listener.openRoot(fileName, supplier, pack);
                        } catch (IOException e) {
                            LibJf.LOGGER.error("Could not call ResourcePack.OPEN_ROOT listener", e);
                            return null;
                        }
                    });
                }
                return lazy.get();
            }));

    public interface Contains {
        boolean contains(ResourceType type, Identifier id, Supplier<Boolean> previous, ResourcePack pack);
    }

    public interface FindResource {
        Collection<Identifier> findResources(ResourceType type, String namespace, String prefix, Predicate<Identifier> allowedPathPredicate, Supplier<Collection<Identifier>> previous, ResourcePack pack);
    }

    public interface Open {
        InputStream open(ResourceType type, Identifier id, Supplier<InputStream> previous, ResourcePack pack) throws IOException;
    }

    public interface OpenRoot {
        InputStream openRoot(String fileName, Supplier<InputStream> previous, ResourcePack pack) throws IOException;
    }
}
